require 'sqlite3'
require 'kconv'
require 'net/http'
Net::HTTP.version_1_2 
require 'rexml/document'
require 'yaml'
require 'time'
require 'digest/sha1'
require 'optparse'

class TencoClient
	# プログラム情報
	PROGRAM_VERSION = '0.07c'
	TRACKRECORD_POST_SIZE = 250   # 一度に送信する対戦結果数
	DUPLICATION_LIMIT_TIME_SECONDS = 2   # タイムスタンプが何秒以内のデータを、重複データとみなすか
	MAIL_ADDRESS_REGEX = /\A[\x01-\x7F]+@(([-a-z0-9]+\.)*[a-z]+|\[\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}\])\z/ # メールアドレスチェック用正規表現
	PLEASE_RETRY_FORCE_INSERT = "<Please Retry in Force-Insert Mode>"  # 強制インサートリトライのお願い文字列
	HTTP_REQUEST_HEADER = {"User-Agent" => "Hikouseki Report Tool #{PROGRAM_VERSION}"}

	# 設定で読み込めなかったときの初期値
	DEFAULT_GAME_ID = 1    # ゲームID
	DEFAULT_DATABASE_FILE_PATH = '../*.db' # データベースファイルパス

	# ログファイルパス
	ERROR_LOG_PATH = 'error.txt'

	def run()
		begin
			result_out = File.open('Tenmite.log', 'w')
		rescue
			return -1
		end
		
		# 送信設定
		is_force_insert = false # 強制インサートモード。はじめは false。
		is_all_report = false # 全件報告モード。サーバーからの最終対戦時刻をとらず、全件送信。

		# 変数
		trackrecord = [] # 対戦結果
		is_warning_exist = false # 警告メッセージがあるかどうか

		result_out.print "*** 緋行跡報告ツール *** ver.#{PROGRAM_VERSION}\n\n".kconv(Kconv::SJIS, Kconv::UTF8)

		begin

			### 設定読み込み ###

			# 設定ファイルパス
			config_file = 'config.yaml'
			config_default_file = 'config_default.yaml'
			env_file = 'env.yaml'

			# 設定ファイルがなければデフォルトをコピーして作成
			unless File.exist?(config_file) then
				open(config_default_file) do |s|
					open(config_file, "w") do |d|
						d.write(s.read)
					end
				end
			end

			# サーバー環境設定ファイルがなければ、エラー終了
			unless File.exist?(env_file) then
				raise "#{env_file} が見つかりません。\nダウンロードした本プログラムのフォルダからコピーしてください。"
			end

			# 設定ファイル読み込み
			# UTF-8 対応のため、一旦 Kconv を通して UTF-8N にする。
			config = YAML.load(File.read(config_file).kconv(Kconv::UTF8, Kconv::UTF8))
			env = YAML.load(File.read(env_file).kconv(Kconv::UTF8, Kconv::UTF8))

			# config.yaml がおかしいと代入時にエラーが出ることに対する格好悪い対策
			config ||= {}
			config['account'] ||= {}
			config['game'] ||= {}
			config['database'] ||= {}

			account_name = config['account']['name'].to_s || ''
			account_password = config['account']['password'].to_s || ''
			game_id = config['game']['id'] || DEFAULT_GAME_ID
			db_file_path = config['database']['file_path'].to_s || DEFAULT_DATABASE_FILE_PATH
															  
			# proxy_host = config['proxy']['host']
			# proxy_port = config['proxy']['port']
			# last_report_time = config['last_report_time']
			# IS_USE_HTTPS = false

			@server_track_record_host = env['server']['track_record']['host'].to_s
			@server_track_record_path = env['server']['track_record']['path'].to_s
			@server_last_track_record_host = env['server']['last_track_record']['host'].to_s
			@server_last_track_record_path = env['server']['last_track_record']['path'].to_s
			@server_account_host = env['server']['account']['host'].to_s
			@server_account_path = env['server']['account']['path'].to_s

			### メソッド定義 ###

			# 試合結果データ読み込み
			# 発生時間が古い順にデータを並べて返す
			# 指定されたデータベースファイルが存在しなければ、例外を発生させる
			# db_file は SJIS で渡すこと
			def read_trackrecord(db_file, last_report_filetime = 0)
				trackrecord = []
				
				# 緋行跡DB接続
				# DBファイルがなければ例外発生
				# 見つからないケースは、UNICODE のファイル名で、うまく取り扱えなかったとき
				raise <<-MSG unless File.exist?(db_file)
緋行跡データベースファイルが見つかりません。(パス：#{db_file.kconv(Kconv::UTF8, Kconv::SJIS)})
・エクスプローラー上で見て、もし該当パス名のファイルがあり、サイズが0KBであれば、削除してください
　「～」などの文字を含むDBファイル名は、サポート対象外です。ファイル名を変更してください。
				MSG

				# DBファイルサイズが0なら処理スキップ
				if File.size?(db_file) == 0
					result_out.puts "！ファイルサイズが 0KB です(パス：#{db_file.kconv(Kconv::UTF8, Kconv::SJIS)})。処理をスキップします。".kconv(Kconv::UTF8, Kconv::SJIS)
					return trackrecord
				end
				
				begin
					begin
						db = SQLite3::Database.open(db_file.kconv(Kconv::UTF8, Kconv::SJIS))
						# SQLite3 モジュールの UTF16 ファイル名対応をONにした場合の読み込み
						# ただ、UTF16対応の必要なファイル名だと、テーブル名が見つからないなど、データが読めないことがある。理由は謎。
						# require 'nkf'
						# db = SQLite3::Database.open(NKF.nkf('-S -w16' ,db_file), {:utf16 => true})
						db.results_as_hash = true
					rescue => ex
						raise <<-MSG
緋行跡データベースファイルへの接続に失敗しました。(パス：#{db_file.kconv(Kconv::UTF8, Kconv::SJIS)})
・エクスプローラー上で見て、もし該当パス名のファイルがあり、サイズが0KBであれば、削除してください
　「～」などの文字を含むDBファイル名は、サポート対象外です。ファイル名を変更してください。
・緋行跡Rev.10**系列を使っているかどうか、確認してください（Rev.1*系列は未対応です）
						MSG
					end
					
					# 試合結果記録の読み込み
					sql = <<-SQL
						SELECT
							*
						FROM
							trackrecord
						WHERE
							    timestamp > #{last_report_filetime.to_i}
							AND	coalesce(p1name, '') != ''
							AND p1id >= 0
							AND p1win >= 0
							AND coalesce(p2name, '') != ''
							AND p2id >= 0
							AND p2win >= 0
						ORDER BY
							timestamp
					SQL

					begin
						db.execute(sql) do |row|
							trackrecord << row
						end
					rescue => ex
						raise <<-MSG
緋行跡データベースへの、対戦結果取得SQLクエリ発行時にエラーが発生しました。(パス：#{db_file.kconv(Kconv::UTF8, Kconv::SJIS)})
・エクスプローラー上で見て、もし該当パス名のファイルがあり、サイズが0KBであれば、削除してください
　「～」などの文字を含むDBファイル名は、サポート対象外です。ファイル名を変更してください。
・緋行跡Rev.10**系列を使っているかどうか、確認してください（Rev.1*系列は未対応です）
						MSG
					end
						
				rescue => ex
					raise ex.to_s
				ensure
					db.close if db
				end
					
				return trackrecord
			end

			# タイムスタンプが一定時間内に連続するデータを削除
			def delete_duplicated_trackrecord(trackrecord)
				last_timestamp = nil    # 直前のデータのタイムスタンプ
				delete_trackrecord = [] # 削除対象のデータ
				
				# タイムスタンプ順にソート
				trackrecord.sort! {|a, b| Time.parse(a['timestamp']) <=> Time.parse(b['timestamp'])}

				# 重複レコードを取得
				trackrecord.each do |t|
					timestamp = Time.parse(t['timestamp'])
					
					unless last_timestamp
						last_timestamp = timestamp
					else
						if timestamp <= last_timestamp + DUPLICATION_LIMIT_TIME_SECONDS
							delete_trackrecord << t
						else
							last_timestamp = timestamp
						end
					end
				end
				
				if delete_trackrecord.length > 0 then
					result_out.puts "#{delete_trackrecord.length} 件のデータは、重複データと疑われるため報告されません。".kconv(Kconv::SJIS, Kconv::UTF8)
					result_out.puts
				end
				
				return trackrecord - delete_trackrecord
			end

			# 緋行跡データをポスト用XMLに変換
			def trackrecord2xml_string(game_id, account_name, account_password, trackrecord, is_force_insert)
				# XML生成
				xml = REXML::Document.new
				xml << REXML::XMLDecl.new('1.0', 'UTF-8')
				
				# trackrecordPosting 要素生成
				root = xml.add_element('trackrecordPosting')
				
				# account 要素生成
				account_element = root.add_element('account')
				account_element.add_element('name').add_text(account_name.to_s)
				account_element.add_element('password').add_text(Digest::SHA1.hexdigest(account_password.to_s))
				
				# game 要素生成
				game_element = root.add_element('game')
				game_element.add_element('id').add_text(game_id.to_s)
				
				# trackrecord 要素生成
				trackrecord.each do |t|
					trackrecord_element = game_element.add_element('trackrecord')
					trackrecord_element.add_element('timestamp').add_text(t['timestamp'].to_s)
					trackrecord_element.add_element('p1name').add_text(t['p1name'].to_s)
					trackrecord_element.add_element('p1type').add_text(t['p1id'].to_s)
					trackrecord_element.add_element('p1point').add_text(t['p1win'].to_s)
					trackrecord_element.add_element('p2name').add_text(t['p2name'].to_s)
					trackrecord_element.add_element('p2type').add_text(t['p2id'].to_s)
					trackrecord_element.add_element('p2point').add_text(t['p2win'].to_s)
				end
				
				# 強制インサート依頼時には、is_force_insert を true にする
				# is_force_insert 要素生成
				if is_force_insert then
					root.add_element('is_force_insert').add_text('true')
				end
				
				return xml.to_s
			end

			# 緋行跡のタイムスタンプ (FILETIME)を ISO8601 形式に変換
			# FILETIME は 1601年1月1日からの100ナノ秒単位での時間
			# 緋行跡はローカル時刻を取得しているので、ローカルの時刻オフセットをつける
			def filetime_to_iso8601 (filetime)
				# DateTime モジュールだと計算精度が低かったので Time モジュールを利用
				if filetime then
					base_filetime = 126227808000000000   # 2001年1月1日0時の FILETIME
					base_time = Time.local(2001, 1, 1)
					time = base_time + (filetime.to_i - base_filetime) / 10.0**7
					return time.iso8601
				else
					return nil
				end
			end

			# Time型のインスタンス を FILETIME に変換
			# FILETIME は 1601年1月1日からの100ナノ秒単位でのローカル時間
			def time_to_filetime (time)
				if time then
				    base_filetime = 126227808000000000   # 2001年1月1日0時の FILETIME
					base_time = Time.local(2001, 1, 1)
					return base_filetime + (time - base_time) * 10**7
				else
					return nil
				end
			end

			# UTF8文字列をデコード
			# YAML(Syck) の to_yaml が日本語対応してない対策
			def decode (string)
				string.gsub(/\\x(\w{2})/){[Regexp.last_match.captures.first.to_i(16)].pack("C")}
			end

			# コンフィグファイル保存
			def save_config (config_file, config)
				File.open(config_file, 'w') do |w|
					w.puts "# 緋行跡報告ツール設定ファイル"
					w.puts "# かならず文字コードは UTF-8 または UTF-8N で保存してください。"
					w.puts "# メモ帳でも編集・保存できます(Ver.0.07c より)。"
					w.puts decode(config.to_yaml)
				end
			end

			### メイン処理 ###

#			【未対応機能】
#			## オプション設定
#			opt = OptionParser.new
#
#			opt.on('-a') {|v| is_all_report = true} # 全件報告モード
#
#			opt.on('--database-filepath-default-overwrite') do |v|  # 設定ファイルのデータベースファイルパスをデフォルトにする
#				result_out.puts "★設定ファイルの緋行跡DBファイルパスを上書きします".kconv(Kconv::SJIS, Kconv::UTF8)
#				result_out.puts "#{config_file} の緋行跡DBファイルパスを #{DEFAULT_DATABASE_FILE_PATH} に書き換え...".kconv(Kconv::SJIS, Kconv::UTF8)
#				config['database']['file_path'] = DEFAULT_DATABASE_FILE_PATH
#				save_config(config_file, config)
#				result_out.puts "設定ファイルを保存しました！".kconv(Kconv::SJIS, Kconv::UTF8)
#				result_out.puts
#				exit
#			end
#
#			opt.parse!(ARGV)

#			【未対応機能】
			## 新規アカウント登録処理
			unless (account_name && account_name =~ /\A[a-zA-Z0-9_\.]+\z/) then
#				account_name = ''
#				account_password = ''
#				is_account_register_finish = false
#				
#				result_out.puts "★新規アカウント登録\n\n".kconv(Kconv::SJIS, Kconv::UTF8)
#				result_out.puts "※もしすでにアカウントを持ってるのであれば、今すぐ Ctrl+C を押して中止し、設定ファイル（#{config_file}）にアカウント名とパスワードを書き込んで、再度実行してください。\n\n".kconv(Kconv::SJIS, Kconv::UTF8)
#
#				while (!is_account_register_finish)
#					# アカウント名入力
#					result_out.puts "希望アカウント名を入力してください\n".kconv(Kconv::SJIS, Kconv::UTF8)
#					result_out.puts "アカウント名はURLの一部として使用されます。\n".kconv(Kconv::SJIS, Kconv::UTF8)
#					result_out.puts "（半角英数とアンダースコア_のみ使用可能。32文字以内）\n".kconv(Kconv::SJIS, Kconv::UTF8)
#					result_out.puts "希望アカウント名：".kconv(Kconv::SJIS, Kconv::UTF8)
#					while (input = gets)
#						input.strip!
#						if input =~ /\A[a-zA-Z0-9_]{1,32}\z/ then
#							account_name = input
#							result_out.puts 
#							break
#						else
#							result_out.puts "！希望アカウント名は半角英数とアンダースコア_のみで、32文字以内で入力してください".kconv(Kconv::SJIS, Kconv::UTF8)
#							result_out.puts "希望アカウント名：".kconv(Kconv::SJIS, Kconv::UTF8)
#						end
#					end
#					
#					# パスワード入力
#					result_out.puts "パスワードを入力してください（使用文字制限なし。4～16byte以内。アカウント名と同一禁止。）\n".kconv(Kconv::SJIS, Kconv::UTF8)
#					result_out.puts "パスワード：".kconv(Kconv::SJIS, Kconv::UTF8)
#					while (input = gets)
#						input.strip!
#						if (input.length >= 4 and input.length <= 16 and input != account_name) then
#							account_password = input
#							break
#						else
#							result_out.puts "！パスワードは4～16byte以内で、アカウント名と別の文字列を入力してください".kconv(Kconv::SJIS, Kconv::UTF8)
#							result_out.puts "パスワード：".kconv(Kconv::SJIS, Kconv::UTF8)
#						end
#					end 
#					
#					result_out.puts "パスワード（確認）：\n".kconv(Kconv::SJIS, Kconv::UTF8)
#					while (input = gets)
#						input.strip!
#						if (account_password == input) then
#							result_out.puts 
#							break
#						else
#							result_out.puts "！パスワードが一致しません\n".kconv(Kconv::SJIS, Kconv::UTF8)
#							result_out.puts "パスワード（確認）：".kconv(Kconv::SJIS, Kconv::UTF8)
#						end
#					end
#					
#					# メールアドレス入力
#					result_out.puts "メールアドレスを入力してください（入力は任意）\n".kconv(Kconv::SJIS, Kconv::UTF8)
#					result_out.puts "※パスワードを忘れたときの連絡用にのみ使用します。\n".kconv(Kconv::SJIS, Kconv::UTF8)
#					result_out.puts "※記入しない場合、パスワードの連絡はできません。\n".kconv(Kconv::SJIS, Kconv::UTF8)
#					result_out.puts "メールアドレス：".kconv(Kconv::SJIS, Kconv::UTF8)
#					while (input = gets)
#						input.strip!
#						if (input == '') then
#							account_mail_address = ''
#							result_out.puts "メールアドレスは登録しません。".kconv(Kconv::SJIS, Kconv::UTF8)
#							result_out.puts
#							break
#						elsif input =~ MAIL_ADDRESS_REGEX and input.length <= 256 then
#							account_mail_address = input
#							result_out.puts
#							break
#						else
#							result_out.puts "！メールアドレスは正しい形式で、256byte以内にて入力してください".kconv(Kconv::SJIS, Kconv::UTF8)
#							result_out.puts "メールアドレス：".kconv(Kconv::SJIS, Kconv::UTF8)
#						end
#					end
#					
#					# 新規アカウントをサーバーに登録
#					result_out.puts "サーバーにアカウントを登録しています...\n".kconv(Kconv::SJIS, Kconv::UTF8)
#					
#					# アカウント XML 生成
#					account_xml = REXML::Document.new
#					account_xml << REXML::XMLDecl.new('1.0', 'UTF-8')
#					account_element = account_xml.add_element("account")
#					account_element.add_element('name').add_text(account_name)
#					account_element.add_element('password').add_text(account_password)
#					account_element.add_element('mail_address').add_text(account_mail_address)
#					# サーバーに送信
#					response = nil
#					# http = Net::HTTP.new(@server_account_host, 443)
#					# http.use_ssl = true
#					# http.verify_mode = OpenSSL::SSL::VERIFY_NONE
#					http = Net::HTTP.new(@server_account_host, 80)
#					http.start do |s|
#						response = s.post(@server_account_path, account_xml.to_s, HTTP_REQUEST_HEADER)
#					end
#					
#					result_out.print "サーバーからのお返事\n".kconv(Kconv::SJIS, Kconv::UTF8)
#					response.body.each_line do |line|
#						result_out.puts "> #{line.kconv(Kconv::SJIS, Kconv::UTF8)}"
#					end
#
#					if response.code == '200' then
#					# アカウント登録成功時
#						is_account_register_finish = true
#						config['account']['name'] = account_name
#						config['account']['password'] = account_password
#						
#						save_config(config_file, config)
#						
#						result_out.puts 
#						result_out.puts "アカウント情報を設定ファイルに保存しました。\n\n".kconv(Kconv::SJIS, Kconv::UTF8)
#						result_out.puts "引き続き、対戦結果の報告をします...\n\n".kconv(Kconv::SJIS, Kconv::UTF8)
#					else
#					# アカウント登録失敗時
#						result_out.puts "もう一度アカウント登録をやり直します...\n\n".kconv(Kconv::SJIS, Kconv::UTF8)
#					end
#					
#					sleep 2
#				end
				result_out.puts "★アカウント情報の取得に失敗しました登録\n\n".kconv(Kconv::SJIS, Kconv::UTF8)
				return -4
			end

			## 【暫定対処】
			## 緋行跡DBが登録済みの最終対戦結果時刻から更新されていない場合は報告しない
			## last_report_trackrecord.xmlが削除されていたりすると上手く判別できない
			if File.exist?('./last_report_trackrecord.xml') then
				lastest_report_time = File.mtime('last_report_trackrecord.xml')
				
				need_report = false
				db_files = Dir::glob(db_file_path.kconv(Kconv::SJIS, Kconv::UTF8))
				if db_files.length > 0
					db_files.each do |db_file|
						if File.mtime(db_file) > lastest_report_time then
							need_report = true
						end
					end
				else
					raise <<-MSG
#{config_file} に設定された緋行跡データベースファイルが見つかりません。
・緋行跡報告ツールのインストール場所が正しいかどうか、確認してください
　デフォルト設定の場合、緋行跡フォルダに、緋行跡報告ツールをフォルダごとおいてください。
・緋行跡 Rev.1* 系列には未対応です。Rev.10** 系列を利用してください。
・#{config_file} を変更した場合、設定が正しいかどうか、確認してください
					MSG
				end
				
				if !need_report then
					result_out.puts "前回の報告から緋行跡DBが更新されていません。\n\n".kconv(Kconv::SJIS, Kconv::UTF8)
					result_out.close
					return 1
				end
			end
			
			
			## 登録済みの最終対戦結果時刻を取得する
			unless is_all_report then
				result_out.puts "★登録済みの最終対戦時刻を取得".kconv(Kconv::SJIS, Kconv::UTF8)
				result_out.puts "GET http://#{@server_last_track_record_host}#{@server_last_track_record_path}?game_id=#{game_id}&account_name=#{account_name}"

				http = Net::HTTP.new(@server_last_track_record_host, 80)
				response = nil
				http.start do |s|
					response = s.get("#{@server_last_track_record_path}?game_id=#{game_id}&account_name=#{account_name}", HTTP_REQUEST_HEADER)
				end

				if response.code == '200' or response.code == '204' then
					if (response.body and response.body != '') then
						last_report_time = Time.parse(response.body)
						result_out.puts "サーバー登録済みの最終対戦時刻：#{last_report_time.strftime('%Y/%m/%d %H:%M:%S')}".kconv(Kconv::SJIS, Kconv::UTF8)
					else
						last_report_time = Time.at(0)
						result_out.puts "サーバーには対戦結果未登録です".kconv(Kconv::SJIS, Kconv::UTF8)
					end
				else
					raise "最終対戦時刻の取得時にサーバーエラーが発生しました。処理を中断します。"
				end
			else
				result_out.puts "★全件報告モードです。サーバーからの登録済み最終対戦時刻の取得をスキップします。".kconv(Kconv::SJIS, Kconv::UTF8)
				last_report_time = Time.at(0)
			end
			result_out.puts

			## 対戦結果報告処理
			result_out.puts "★対戦結果送信".kconv(Kconv::SJIS, Kconv::UTF8)
			result_out.puts( ("緋行跡の記録から、" + last_report_time.strftime('%Y/%m/%d %H:%M:%S') + " 以降の対戦結果を報告します。").kconv(Kconv::SJIS, Kconv::UTF8))
			result_out.puts

			# 緋行跡DBから対戦結果を取得
			db_files = Dir::glob(db_file_path.kconv(Kconv::SJIS, Kconv::UTF8))

			if db_files.length > 0
				db_files.each do |db_file|
					result_out.puts db_file + " から対戦結果を抽出...\n".kconv(Kconv::SJIS, Kconv::UTF8)
					begin
						trackrecord.concat(read_trackrecord(db_file, time_to_filetime(last_report_time + 1)))
					rescue => ex
						is_warning_exist = true
						result_out.puts "！警告".kconv(Kconv::SJIS, Kconv::UTF8)
						result_out.puts ex.to_s.kconv(Kconv::SJIS, Kconv::UTF8)
						result_out.puts "処理を続行します...".kconv(Kconv::SJIS, Kconv::UTF8)
						result_out.puts
					end
				end
			else
				raise <<-MSG
#{config_file} に設定された緋行跡データベースファイルが見つかりません。
・緋行跡報告ツールのインストール場所が正しいかどうか、確認してください
　デフォルト設定の場合、緋行跡フォルダに、緋行跡報告ツールをフォルダごとおいてください。
・緋行跡 Rev.1* 系列には未対応です。Rev.10** 系列を利用してください。
・#{config_file} を変更した場合、設定が正しいかどうか、確認してください
				MSG
			end

			result_out.puts

			## 報告対象データのデータ形式変換・文字コード変換・重複削除

			# タイムスタンプをFILETIMEからISO8601形式に変換
			trackrecord.each do |t|
				t['timestamp'] = filetime_to_iso8601(t['timestamp'])
			end

			# もしタイムスタンプが一定時時間以内のデータがあれば、古いほうを残して報告対象からはずす
			# 1行1行のハッシュを別レコードとして扱うため、連番をふる
			trackrecord.each_index do |i|
				trackrecord[i]['seq'] = i
			end
			trackrecord = delete_duplicated_trackrecord(trackrecord)

			# 文字列をutf-8に変換
			trackrecord.each do |t|
				t['p1name'] = t['p1name'].kconv(Kconv::UTF8, Kconv::SJIS)
				t['p2name'] = t['p2name'].kconv(Kconv::UTF8, Kconv::SJIS)
			end

			## 報告対象データの送信処理

			# 報告対象データが0件なら送信しない
			if trackrecord.length <= 0 then
				result_out.puts "報告対象データはありませんでした。".kconv(Kconv::SJIS, Kconv::UTF8)
			else
				
				# 対戦結果データを分割して送信
				0.step(trackrecord.length, TRACKRECORD_POST_SIZE) do |start_row_num|
					end_row_num = [start_row_num + TRACKRECORD_POST_SIZE - 1, trackrecord.length - 1].min
					response = nil # サーバーからのレスポンスデータ
					
					result_out.puts "#{trackrecord.length}件中の#{start_row_num + 1}件目～#{end_row_num + 1}件目を送信しています#{is_force_insert ? "（強制インサートモード）" : ""}...\n".kconv(Kconv::SJIS, Kconv::UTF8)
					
					# 送信用XML生成
					trackrecord_xml_string = trackrecord2xml_string(game_id, account_name, account_password, trackrecord[start_row_num..end_row_num], is_force_insert)
					File.open('./last_report_trackrecord.xml', 'w') do |w|
						w.puts trackrecord_xml_string
					end

					# データ送信
					# https = Net::HTTP.new(@server_track_record_host, 443)
					# https.use_ssl = true
					# https.verify_mode = OpenSSL::SSL::VERIFY_NONE
					# https = Net::HTTP::Proxy(proxy_addr, proxy_port).new(@server_track_record_host,443)
					# https.ca_file = '/usr/share/ssl/cert.pem'
					# https.verify_depth = 5
					# https.verify_mode = OpenSSL::SSL::VERIFY_PEER
					http = Net::HTTP.new(@server_track_record_host, 80)
					http.start do |s|
						response = s.post(@server_track_record_path, trackrecord_xml_string, HTTP_REQUEST_HEADER)
					end
					
					# 送信結果表示
					result_out.puts "サーバーからのお返事".kconv(Kconv::SJIS, Kconv::UTF8)
					response.body.each_line do |line|
						result_out.puts "> #{line.kconv(Kconv::SJIS, Kconv::UTF8)}"
					end
					result_out.puts
					
					if response.code == '200' then
						sleep 1
						# 特に表示しない
					else
						if response.body.index(PLEASE_RETRY_FORCE_INSERT)
							result_out.puts "強制インサートモードで報告しなおします。5秒後に報告再開...\n\n".kconv(Kconv::SJIS, Kconv::UTF8)
							sleep 5
							is_force_insert = true
							redo
						else
							raise "報告時にサーバー側でエラーが発生しました。処理を中断します。"
						end
					end
				end
			end

			# 設定ファイル更新
			save_config(config_file, config)
					
			result_out.puts

			# 終了メッセージ出力
			if is_warning_exist then
				result_out.puts "報告処理は正常に終了しましたが、警告メッセージがあります。".kconv(Kconv::SJIS, Kconv::UTF8)
				result_out.puts "出力結果をご確認ください。".kconv(Kconv::SJIS, Kconv::UTF8)
				result_out.puts
#				result_out.puts "Enter キーを押すと、処理を終了します。".kconv(Kconv::SJIS, Kconv::UTF8)
#				exit if gets
				result_out.puts
				return -2
			else
				result_out.puts "報告処理が正常に終了しました。".kconv(Kconv::SJIS, Kconv::UTF8)
			end

#			sleep 3
			result_out.close
			return 0

			### 全体エラー処理 ###
		rescue => ex
			if config && config['account'] then
				config['account']['name']     = '<secret>' if config['account']['name']
				config['account']['password'] = '<secret>' if config['account']['password']
			end
			
			result_out.puts 
			result_out.puts "処理中にエラーが発生しました。処理を中断します。\n".kconv(Kconv::SJIS, Kconv::UTF8)
			result_out.puts 
			result_out.puts '### エラー詳細ここから ###'.kconv(Kconv::SJIS, Kconv::UTF8)
			result_out.puts
			result_out.puts ex.to_s.kconv(Kconv::SJIS, Kconv::UTF8)
			result_out.puts
			result_out.puts ex.backtrace.join("\n").kconv(Kconv::SJIS, Kconv::UTF8)
			result_out.puts( (config ? decode(config.to_yaml) : "config が設定されていません。").kconv(Kconv::SJIS, Kconv::UTF8))
			if response then
				result_out.puts
				result_out.puts "<サーバーからの最後のメッセージ>".kconv(Kconv::SJIS, Kconv::UTF8)
				result_out.puts "HTTP status code : #{response.code}"
				result_out.puts response.body.kconv(Kconv::SJIS, Kconv::UTF8)
			end
			result_out.puts
			result_out.puts '### エラー詳細ここまで ###'.kconv(Kconv::SJIS, Kconv::UTF8)
			
			File.open(ERROR_LOG_PATH, 'a') do |log|
				log.puts "#{Time.now.strftime('%Y/%m/%d %H:%M:%S')} #{File::basename(__FILE__)} #{PROGRAM_VERSION}" 
				log.puts ex.to_s
				log.puts ex.backtrace.join("\n")
				log.puts config ? decode(config.to_yaml) : "config が設定されていません。"
				if response then
					log.puts "<サーバーからの最後のメッセージ>"
					log.puts "HTTP status code : #{response.code}"
					log.puts response.body
				end
				log.puts '********'
			end
			
			result_out.puts
			result_out.puts "上記のエラー内容を #{ERROR_LOG_PATH} に書き出しました。".kconv(Kconv::SJIS, Kconv::UTF8)
			result_out.puts
			
#			result_out.puts "Enter キーを押すと、処理を終了します。".kconv(Kconv::SJIS, Kconv::UTF8)
#			exit if gets
			result_out.close
			return -3
		end
	end
end

# test code
TencoClient.new.run
